use super::*;
use crate::cache::Cache;
use crate::model::{
    Error, Result, auth::User, journal::JournalPageMembership,
    journal_permissions::JournalPermission, permissions::FinePermission,
};
use crate::{auto_method, execute, get, query_row};

#[cfg(feature = "sqlite")]
use rusqlite::Row;

#[cfg(feature = "postgres")]
use tokio_postgres::Row;

impl DataManager {
    /// Get a [`JournalEntry`] from an SQL row.
    pub(crate) fn get_membership_from_row(
        #[cfg(feature = "sqlite")] x: &Row<'_>,
        #[cfg(feature = "postgres")] x: &Row,
    ) -> JournalPageMembership {
        JournalPageMembership {
            id: get!(x->0(u64)) as usize,
            created: get!(x->1(u64)) as usize,
            owner: get!(x->2(u64)) as usize,
            journal: get!(x->3(u64)) as usize,
            role: JournalPermission::from_bits(get!(x->4(u32))).unwrap(),
        }
    }

    auto_method!(get_membership_by_id()@get_membership_from_row -> "SELECT * FROM memberships WHERE id = $1" --name="journal membership" --returns=JournalPageMembership --cache-key-tmpl="atto.membership:{}");

    /// Get a journal page membership by `owner` and `journal`.
    pub async fn get_membership_by_owner_journal(
        &self,
        owner: usize,
        journal: usize,
    ) -> Result<JournalPageMembership> {
        let conn = match self.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = query_row!(
            &conn,
            "SELECT * FROM memberships WHERE owner = $1 AND journal = $2",
            &[&owner, &journal],
            |x| { Ok(Self::get_membership_from_row(x)) }
        );

        if res.is_err() {
            return Err(Error::GeneralNotFound("journal membership".to_string()));
        }

        Ok(res.unwrap())
    }

    /// Create a new journal page membership in the database.
    ///
    /// # Arguments
    /// * `data` - a mock [`JournalPageMembership`] object to insert
    pub async fn create_membership(&self, data: JournalPageMembership) -> Result<()> {
        let conn = match self.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(
            &conn,
            "INSERT INTO memberships VALUES ($1, $2, $3, $4, $5",
            &[
                &data.id.to_string().as_str(),
                &data.created.to_string().as_str(),
                &data.owner.to_string().as_str(),
                &data.journal.to_string().as_str(),
                &(data.role.bits()).to_string().as_str(),
            ]
        );

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        Ok(())
    }

    auto_method!(delete_membership()@get_membership_by_id:MANAGE_MEMBERSHIPS -> "DELETE FROM memberships WHERE id = $1" --cache-key-tmpl="atto.membership:{}");
}
